#ifndef MANAGERS_H
#define MANAGERS_H

#include "global.h"
#include "sprites.h"
#include <list>
#include <utility>
#include <vector>


// An attack wave is defined by the alien vessel, the number of
// vessels and the attack formation (that can be X-mirrored)
struct WaveData {
	int wait;
	const char ** alien_name;
	int wave_size_pc;
	const RCoord * arrival;
	size_t n_rcoords;
	bool mirrored;
};

class Wave {
	Wave(const Wave &); // no copy
	Wave & operator=(const Wave &); // no copy

	const Pix pix_;
	const int wait_;
	const int wave_size_pc_;
	const std::vector<RCoord> arrival_;
	const bool mirrored_;
public:
	Wave(const WaveData &);
	int Wait() const { return wait_; }
	const Pix * AlienPix() const { return &pix_; }
	int Size() const;
	const std::vector<RCoord> & Arrival() const { return arrival_; }
	bool Mirrored() const { return mirrored_; }
};


// Aliens arrive in waves. The AlienManager creates aliens according
// to waves data and then move and destroy them if collision.
class AliensManager {
	static int bonus_wait_;

	typedef std::vector<const Wave*> WavesCtr;
	WavesCtr waves_;
	unsigned wave_number_;
	int wave_alien_number_;
	int speed_;
	Coord base_cruise_;
	int base_cruise_speed_;
	enum creation_t { new_wave, new_alien, new_attack };
	creation_t next_creation_;
	int next_creation_wait_;
	int inter_aliens_wait_;
	int attack_wait_;
	int nb_attacking_max_;
	int inter_bonus_wait_;
	typedef std::list<Alien*> AliensCtr;
	AliensCtr aliens_;
public:
	AliensManager(int level_number, const WaveData * wavesdata,
	              int nwaves, int max_level);
	~AliensManager();
	void Move();
	std::list<Bullet*> Fire(int max_nb) const;
	std::pair<Explosion*, Bonus*> Collision(const Bullet*);
	bool Finished() const;
private:
	void Creation();
	Bonus * CreateBonusMaybe(Coord);
};


class BulletsManager {
public:
	typedef std::list<Bullet*> BulletsCtr;
private:
	BulletsCtr bullets_;
public:
	BulletsManager() {}
	~BulletsManager();
	void Move();
	void AddIf(Bullet * b) { if (b) bullets_.push_back(b); }
	void Add(BulletsCtr bl)
		{ bullets_.splice(bullets_.end(), bl); }
	BulletsCtr & GetList()  { return bullets_; }
	BulletsCtr::iterator Destroy(BulletsCtr::iterator);
	size_t Nb() const { return bullets_.size(); }
};


class BonusManager {
public:
	typedef std::list<Bonus*> BonusCtr;
private:
	BonusCtr bonuses_;
public:
	BonusManager() {}
	~BonusManager();
	void Move();
	void AddIf(Bonus * b) { if (b) bonuses_.push_back(b); }
	BonusCtr & GetList() { return bonuses_; }
	BonusCtr::iterator Destroy(BonusCtr::iterator);
};


class ExplosionsManager {
	typedef std::list<Explosion*> ExplosionsCtr;
	ExplosionsCtr explosions_;
public:
	ExplosionsManager() {}
	~ExplosionsManager();
	void Move();
	void Add(Explosion * e) { explosions_.push_back(e); }
};


#endif
